import argparse
from util.normalize import *
#TODO: change the paths of libsvm and yeal tools in pyaction module
from util.pyaction import *
import pdb

# This file provide the baseline experiment on the HMDB51 dataset with
# Improved dense trajectories and Fisher vector.
# You may need to download video data for HMDB51 dataset.
#
# Note : Please check TODOs : change the paths etc..
# 
# Dependency : 
# 1. opencv-2.4.9
# 2. IDT (modified version which saves IDT in a faster scheme included in 'bin/')
#
# Version      : 0.1.0
# Release date : 2015/09/12
# Author       : Xiaojiang Peng | Email: xiaojiangp@gmail.com
# 
# If you find this useful, please refer to:
# Peng, X., Wang, L., Wang, X., Qiao, Y.: Bag of visual words and fusion methods for
#   action recognition: Comprehensive study and good practice. CoRR abs/1405.4506.(2014)


def get_config(split):
  #TODO : Change the paths
  fv_dir = '/scratch/gpuhost2/xpeng/data/HMDB51-VTE-LSTM/fvecs_s' + str(split) 
  vocab_dir = '/scratch/gpuhost2/xpeng/data/HMDB51-VTE-LSTM/~temp'
  desc_dir = '/home/lear/xpeng/data_scratch2/hmdb51_org_idt'
  video_dir = '/home/lear/xpeng/data_scratch2/hmdb51_org_idt'
  splitdir = '/scratch/gpuhost2/xpeng/data/HMDB51-VTE-LSTM/HMDB51_TestTrain_7030_splits'
  videonames, classlabels, tr_index, actions= get_hmdb_split(split,splitdir)
  return videonames, classlabels, fv_dir, vocab_dir, desc_dir, video_dir, actions, tr_index

def run_hmdb51_split(args):
  parser = argparse.ArgumentParser(
      description="Baseline experiments with Improved dense trajectories and Fisher vector.")

  parser.add_argument(
      '--descriptor', choices=['hog', 'hof', 'mbhx', 'mbhy'], nargs='+', default = ['hog','hof','mbhx','mbhy'],
      help="what features to use.")
  parser.add_argument(
      '--fusion', choices=['d-fusion', 'r-fusion', 's-fusion'], default='r-fusion', help="type of fusion.")
  parser.add_argument(
      '--split', type=int, default=1, help="split number")
  parser.add_argument(
      '--encode', choices=['fv','svc','svc-k','svc-all','vlad','vlad-k','vlad-all','llc','sa-k','vq'],
      default='fv', help="which encoding method to use")
  parser.add_argument(
      '--normalize', choices=['power_l2norm', 'power_intra_l2norm'], default='power_l2norm', help="normalization method")
  parser.add_argument(
      '--gmmsize', type=int, default= 256, help=" GMM size")
  
  args = parser.parse_args()
  
  assert 1 <= args.split <= 3, "Invalid split number."
  
  #TODO: change the paths in get_config function
  videonames, classlabels, fv_dir, vocab_dir, desc_dir, video_dir, actions, tr_index = get_config(args.split)
  
  #extract improved dense trajectories
  extract_idt(video_dir, videonames, desc_dir)
  
  #generate gmm 
  gmmFilePath = os.path.join(vocab_dir,'%d_gmmModel_%d.pkl' % (args.split,args.gmmsize))
  if os.path.exists(gmmFilePath):
    encoderobj = open(gmmFilePath,'rb')
    encoder = pickle.load(encoderobj)
  else:
    encoder = get_idt_encoder(args.split, [videonames[i] for i, x in enumerate(tr_index) if x], vocab_dir, desc_dir, args.gmmsize)    
    output = open(gmmFilePath, 'wb')
    pickle.dump(encoder, output , -1)
  
  #encode features
  fv_savepath = os.path.join(fv_dir,'%d_fv_feat_all.npz' % args.split)
  if os.path.exists(fv_savepath):
    tmp = np.load(fv_savepath)
    feat_all = (tmp['fvecs_hog'], tmp['fvecs_hof'], tmp['fvecs_mbhx'], tmp['fvecs_mbhy'])
  else:
    feat_all = encode_videos(videonames, encoder, fv_dir, desc_dir, args.encode)
    np.savez(fv_savepath, fvecs_hog=feat_all[0], fvecs_hof=feat_all[1], fvecs_mbhx=feat_all[2], fvecs_mbhy=feat_all[3])
  
  #classify: for simplity, only r-fusion is supported for fusion.
  train_labels = [classlabels[i] for i, x in enumerate(tr_index) if x ]
  ts_index = [not x for x in tr_index]
  test_labels = [classlabels[i] for i, x in enumerate(tr_index) if not x ]
  tr_kern_sum = np.zeros((len(train_labels),len(train_labels)),dtype=np.float32)
  ts_kern_sum = np.zeros((len(train_labels),len(test_labels)), dtype=np.float32)
  desc_acc = np.zeros((len(args.descriptor)+1,1),dtype=np.float32)
  for i in range(len(args.descriptor)):
    feature = feat_all[i]
    feature = normalize(feature, args.normalize, 2*args.gmmsize)
    tr_data = feature[np.asarray(tr_index),:]
    ts_data = feature[np.asarray(ts_index),:]
    tr_kern = np.dot(tr_data, tr_data.T)
    ts_kern = np.dot(tr_data, ts_data.T)
    tr_kern_sum = tr_kern_sum + tr_kern
    ts_kern_sum = ts_kern_sum + ts_kern
    avg_acc, cm = classify(tr_kern, ts_kern, train_labels, test_labels, 51)
    print '%s accuracy: %f ' % (args.descriptor[i], avg_acc)
    desc_acc[i] = avg_acc
  avg_acc, cm = classify(tr_kern_sum, ts_kern_sum, train_labels, test_labels, 51)
  print '%d descs combination accuracy: %f ' % (len(args.descriptor), avg_acc)
  desc_acc[len(args.descriptor)] = avg_acc
  return desc_acc
  
def main():
  desc_accs = np.zeros((5,4),dtype=np.float32)
  for i in range(3):
    desc_accs[:,i] = run_hmdb51_split('--split ' + str(i+1))
  desc_accs[:, 3] = np.mean(desc_accs[:,:3], axis=1)
  print desc_accs

if __name__ == "__main__":
  main()
  
  
  
