import os
import sys
import pickle
import numpy as np
#TODO: change the following path
sys.path.append('/home/lear/xpeng/software/toolkit/yael')
sys.path.append('/home/lear/xpeng/software/toolkit/libsvm-3.14/python')
from yael import yael
from yael import ynumpy
from svmutil import *

def get_cm(gt,predict_label,FLAG):
  nr_label = gt.shape[0]
  ncls = np.amax(np.unique(gt))
  cm = np.zeros((ncls,ncls),dtype=np.float32)
  for i in range(nr_label):
    cm[gt[i]-1,predict_label[i]-1] = cm[gt[i]-1,predict_label[i]-1] + 1
  tp = np.trace(cm)
  avg_acc = tp/nr_label
  if FLAG:
    sums = np.sum(cm, axis=1)
    for i in range(ncls):
      if sums[i]!=0:
        cm[i, :] = cm[i, :] / sums[i]
  return avg_acc, cm

def classify(tr_kern, ts_kern, train_labels, test_labels, ncls):
  score_test = svm_one_vs_all(tr_kern, ts_kern, train_labels, ncls)
  predict_labels = np.argmax(score_test, axis=1) + 1
  avg_acc, cm = get_cm(np.asarray(test_labels), predict_labels, True)
  return avg_acc, cm

def xpca(X, whiten, K):
  # X is in row-wise
  d = {}
  dim = X.shape[1]
  whiteningRegul = 0.01;
  mean = X.mean(axis = 0)
  X = X - mean
  cov = np.dot(X.T, X)
  eigvals, eigvecs = np.linalg.eig(cov)
  perm = eigvals.argsort()[::-1][:dim] # sort by decending eigenvalue  
  eigvals = eigvals + whiteningRegul * eigvals[perm[0]]
  if whiten==1:
    pcamap = np.dot(np.diag(1/np.sqrt(eigvals[:K])), eigvecs[:, perm[:K]].T)
    pcamap = pcamap.T
  else:
    pcamap = eigvecs[:, perm[:K]]
  d['pcamap'] = pcamap
  d['center'] = mean
  # pcamap: (n,K)  mean: n,
  return d

def read_idt_bin(filepath):
  index = [[1,41,137,245,341],[40,136,244,340,436]]
  #index = [[1,38,134,242,338],[37,133,241,337,433]]
  sizef = os.path.getsize(filepath)
  if sizef!=0:
    video_dt = np.fromfile(filepath,dtype=np.float32)
    if video_dt.shape[0]%index[1][4]!=0:
      dt = 'feature dimension does not match'
    else:
      dt = np.reshape(video_dt, (-1,index[1][4]))
  else:
    dt = 'No feature'
  return dt, index

def extract_improvedfeatures(videofile,outfile):
  command = './bin/DenseTrackStab %s -o %s' % (videofile,outfile)
  os.system(command)
  
def extract_idt(video_dir, videonames, desc_dir):
  if not os.path.exists(desc_dir):
    os.mkdir(desc_dir)
  parts = videonames[0].split('/')  
  if not os.path.exists(os.path.join(desc_dir,parts[0])):
    for i in range(len(videonames)):
      parts = videonames[i].split('/')
      mkpath = os.path.join(desc_dir,parts[0])
      if not os.path.exists(mkpath):
        os.mkdir(mkpath)
  for i in range(len(videonames)):
    descriptor_file = os.path.join(desc_dir, videonames[i]+'.bin')
    if not os.path.exists(descriptor_file):
      videopath = '%s/%s.avi' % (video_dir, videonames[i])
      extract_improvedfeatures(videopath,descriptor_file)
    sizef = os.path.getsize(descriptor_file)/(1024*1024)
    print '%d -> %s --> size %f Mb' % (i,descriptor_file,sizef)

def get_hmdb_split(split,splitdir):
  saction = ['brush_hair','cartwheel','catch','chew','clap','climb','climb_stairs',\
'dive','draw_sword','dribble','drink','eat','fall_floor','fencing',\
'flic_flac','golf','handstand','hit','hug','jump','kick',\
'kick_ball','kiss','laugh','pick','pour','pullup','punch',\
'push','pushup','ride_bike','ride_horse','run','shake_hands','shoot_ball',\
'shoot_bow','shoot_gun','sit','situp','smile','smoke','somersault',\
'stand','swing_baseball','sword','sword_exercise','talk','throw','turn',\
'walk','wave']  
  video_names = []
  class_labels = []
  tr_index = []  
  for iaction in range(len(saction)):
    fname = splitdir + '/' + saction[iaction] + '_test_split' + str(split) + '.txt'
    with open(fname,'r') as f:
      for line in f:
        sline = line.rstrip('\n').split(' ')
        u = int(sline[1])
        video = sline[0].replace('.avi', '')
        if u==1: # train
          video_names.append('%s/%s' % (saction[iaction], video))
          class_labels.append(iaction+1)
          tr_index.append(True)
        elif u==2:
          video_names.append('%s/%s' % (saction[iaction], video))
          class_labels.append(iaction+1)
          tr_index.append(False)
  return video_names, class_labels, tr_index, saction

def svm_one_vs_all(tr_kern, ts_kern, train_labels, ncls):
  score_test = np.zeros((ts_kern.shape[1], ncls), dtype=np.float32)
  N = tr_kern.shape[0]
  class_train = np.asarray(train_labels) -1
  for class_ind in range(ncls):
    print class_ind
    Y = np.zeros((N,1), dtype=np.int16)-1
    Y[class_train == class_ind] = 1
    X = [None] * N
    for i in range(N):
      X[i] = {j+1:tr_kern[i, j] for j in range(N)}
      X[i][0] = i+1
    prob  = svm_problem(Y, X, isKernel=True)
    param = svm_parameter('-t 4 -c 100')
    model = svm_train(prob,param)
    rho = model.rho[0] 
    sv_coeffs = np.asarray(model.get_sv_coef())
    sv_indices = np.asarray(model.get_sv_indices()) - 1
    if Y[0]==-1:
      sv_coeffs = -sv_coeffs
      rho = -rho
    score_test[:,class_ind] = np.dot(sv_coeffs.T, ts_kern[sv_indices, :]) - rho
  return score_test
  
def encode_videos(videonames, gmm, fv_dir, desc_dir, encode):
  if not os.path.exists(fv_dir):
    os.mkdir(fv_dir)
  parts = videonames[0].split('/')  
  if not os.path.exists(os.path.join(fv_dir,parts[0])):
    for i in range(len(videonames)):
      parts = videonames[i].split('/')
      mkpath = os.path.join(fv_dir,parts[0])
      if not os.path.exists(mkpath):
        os.mkdir(mkpath)
  fv_hog = np.zeros((len(videonames),gmm['pca']['hog']['pcamap'].shape[1]*2*gmm['gmm']['hog'][0].shape[0]), dtype=np.float32)
  fv_hof = np.zeros((len(videonames),gmm['pca']['hof']['pcamap'].shape[1]*2*gmm['gmm']['hof'][0].shape[0]), dtype=np.float32)
  fv_mbhx = np.zeros((len(videonames),gmm['pca']['mbhx']['pcamap'].shape[1]*2*gmm['gmm']['mbhx'][0].shape[0]), dtype=np.float32)
  fv_mbhy = np.zeros((len(videonames),gmm['pca']['mbhy']['pcamap'].shape[1]*2*gmm['gmm']['mbhy'][0].shape[0]), dtype=np.float32)
  for i in range(len(videonames)):
    print '%d/%d -> %s ' % (i,len(videonames),videonames[i])
    savefile = os.path.join(fv_dir, '%s.npz' % videonames[i]);
    if not os.path.exists(savefile):
      descriptorFile = os.path.join(desc_dir,'%s.bin' % videonames[i])
      dt, index = read_idt_bin(descriptorFile)
      if type(dt) is str:
        print 'no feature or feature error with %s' % descriptorFile
        fv_hog[i,:] = 1/fv_hog.shape[1]
        fv_hof[i,:] = 1/fv_hof.shape[1]
        fv_mbhx[i,:] = 1/fv_mbhx.shape[1]
        fv_mbhy[i,:] = 1/fv_mbhy.shape[1]
      else:
        # hog
        desc = dt[:,(index[0][1]-1):index[1][1]]
        desc = np.dot(desc - gmm['pca']['hog']['center'], gmm['pca']['hog']['pcamap'])
        desc = np.copy(desc, order='C')
        fv_hog[i,:]= ynumpy.fisher(gmm['gmm']['hog'], desc, include =['mu', 'sigma'])
        # hof
        desc = dt[:,(index[0][2]-1):index[1][2]]
        desc = np.dot(desc - gmm['pca']['hof']['center'], gmm['pca']['hof']['pcamap'])
        desc = np.copy(desc, order='C')
        fv_hof[i,:]= ynumpy.fisher(gmm['gmm']['hof'], desc, include =['mu', 'sigma'])
        # mbhx
        desc = dt[:,(index[0][3]-1):index[1][3]]
        desc = np.dot(desc - gmm['pca']['mbhx']['center'], gmm['pca']['mbhx']['pcamap'])
        desc = np.copy(desc, order='C')
        fv_mbhx[i,:]= ynumpy.fisher(gmm['gmm']['mbhx'], desc, include =['mu', 'sigma'])
        # mbhy
        desc = dt[:,(index[0][4]-1):index[1][4]]
        desc = np.dot(desc - gmm['pca']['mbhy']['center'], gmm['pca']['mbhy']['pcamap'])
        desc = np.copy(desc, order='C')
        fv_mbhy[i,:]= ynumpy.fisher(gmm['gmm']['mbhy'], desc, include =['mu', 'sigma'])
      np.savez(savefile, fvec_hog=fv_hog[i,:],fvec_hof=fv_hof[i,:],fvec_mbhx=fv_mbhx[i,:],fvec_mbhy=fv_mbhy[i,:])
    else:
      tmp = np.load(savefile)
      fv_hog[i,:] = tmp['fvec_hog'] 
      fv_hof[i,:] = tmp['fvec_hof']
      fv_mbhx[i,:] = tmp['fvec_mbhx']
      fv_mbhy[i,:] = tmp['fvec_mbhy']
    
  return fv_hog, fv_hof, fv_mbhx, fv_mbhy

def get_idt_encoder(split, videonames, vocab_dir, desc_dir, gmmsize):
  samples = 256000
  pcaFactor = 0.5
  whiten = 1
  sample_feat_file = os.path.join(vocab_dir,'%d_featfile.npz' % split)
  encoder = {}
  encoder['gmm'] = {}
  encoder['pca'] = {}
  if not os.path.exists(sample_feat_file):
    hogAll = np.zeros((samples,96),dtype=np.float32)
    hofAll = np.zeros((samples,108),dtype=np.float32)
    mbhxAll = np.zeros((samples,96),dtype=np.float32)
    mbhyAll = np.zeros((samples,96),dtype=np.float32)
    print 'getEncoder : generate encoder from subset of videos...'
    num_videos = np.minimum(3000, len(videonames))
    num_samples_per_vid = np.ceil(samples/ num_videos)
    
    videoname = [videonames[i] for i in np.random.permutation(len(videonames))]
    st = 0
    for i in range(num_videos):
      print '%d/%d -> %s ' % (i,num_videos,videoname[i])
      descriptorFile = os.path.join(desc_dir,'%s.bin' % videoname[i])
      dt, index = read_idt_bin(descriptorFile)
      if type(dt) is str:
        print 'no feature or feature error with %s' % descriptorFile
      else:
        rnsam = np.random.permutation(dt.shape[0])
        if dt.shape[0] > num_samples_per_vid:
          rnsam = rnsam[:num_samples_per_vid]
        send = st + rnsam.shape[0]
        hogAll[st:send,:] = dt[rnsam,(index[0][1]-1):index[1][1]]
        hofAll[st:send,:] = dt[rnsam,(index[0][2]-1):index[1][2]]
        mbhxAll[st:send,:] = dt[rnsam,(index[0][3]-1):index[1][3]]
        mbhyAll[st:send,:] = dt[rnsam,(index[0][4]-1):index[1][4]]
      st = st + rnsam.shape[0]
    np.savez(sample_feat_file, hog=hogAll, hof=hofAll, mbhx=mbhxAll, mbhy=mbhyAll)
  else:
    tmp = np.load(sample_feat_file)
    hogAll = tmp['hog']
    hofAll = tmp['hof']
    mbhxAll = tmp['mbhx']
    mbhyAll = tmp['mbhy']
  print 'start computing pca'
  pcahog = xpca(hogAll, whiten, hogAll.shape[1]*pcaFactor)
  encoder['pca']['hog'] = pcahog
  pcahof = xpca(hofAll, whiten, hofAll.shape[1]*pcaFactor)
  encoder['pca']['hof'] = pcahof
  pcambhx = xpca(mbhxAll, whiten, mbhxAll.shape[1]*pcaFactor)
  encoder['pca']['mbhx'] = pcambhx
  pcambhy = xpca(mbhyAll, whiten, mbhyAll.shape[1]*pcaFactor)
  encoder['pca']['mbhy'] = pcambhy  
  print 'start learning gmm for hog'
  hogAll = hogAll - pcahog['center']
  hogAll = np.dot(hogAll, pcahog['pcamap'])
  gmmhog = ynumpy.gmm_learn(hogAll, gmmsize)
  encoder['gmm']['hog'] = gmmhog
  print 'start learning gmm for hof'
  hofAll = hofAll - pcahof['center']
  hofAll = np.dot(hofAll, pcahof['pcamap'])
  gmmhof = ynumpy.gmm_learn(hofAll, gmmsize)
  encoder['gmm']['hof'] = gmmhof
  print 'start learning gmm for mbhx'
  mbhxAll = mbhxAll - pcambhx['center']
  mbhxAll = np.dot(mbhxAll, pcambhx['pcamap'])
  gmmmbhx = ynumpy.gmm_learn(mbhxAll, gmmsize)
  encoder['gmm']['mbhx'] = gmmmbhx
  print 'start learning gmm for mbhy'
  mbhyAll = mbhyAll - pcambhy['center']
  mbhyAll = np.dot(mbhyAll, pcambhy['pcamap'])
  gmmmbhy = ynumpy.gmm_learn(mbhyAll, gmmsize)
  encoder['gmm']['mbhy'] = gmmmbhy
  return encoder
  

